@@ -324,6 +324,14 @@ public class Main {
             return;
         }
 
+        // Keep support for older (-version/--version) command line arguments
+        for (String arg : args) {
+            if (arg.equals("-version") || arg.equals("--version")) {
+                printVersion();
+                return;
+            }
+        }
+
         String cmdName = args[0];
         String[] cmdArgs = Arrays.copyOfRange(args, 1, args.length);
 
