@@ -16,8 +16,6 @@
  */
 package brut.androlib.res.data.value;
 
-import brut.androlib.exceptions.AndrolibException;
-
 public class ResFileValue extends ResIntBasedValue {
     private final String mPath;
 
@@ -26,16 +24,6 @@ public class ResFileValue extends ResIntBasedValue {
         mPath = path;
     }
 
-    public String getStrippedPath() throws AndrolibException {
-        if (mPath.startsWith("res/")) {
-            return mPath.substring(4);
-        }
-        if (mPath.startsWith("r/") || mPath.startsWith("R/")) {
-            return mPath.substring(2);
-        }
-        throw new AndrolibException("File path does not start with \"res/\" or \"r/\": " + mPath);
-    }
-
     @Override
     public String toString() {
         return mPath;
