@@ -18,8 +18,13 @@ package brut.xmlpull;
 
 import org.xmlpull.v1.XmlSerializer;
 
-import java.io.*;
-import java.util.*;
+import java.io.IOException;
+import java.io.OutputStream;
+import java.io.OutputStreamWriter;
+import java.io.Writer;
+import java.util.HashSet;
+import java.util.Objects;
+import java.util.Set;
 
 /**
  * Implementation of XmlSerializer interface from XmlPull V1 API. This
@@ -1103,7 +1108,7 @@ public class MXSerializer implements XmlSerializer {
                 retval.append("\\\\");
                 break;
             default:
-                if (ch < 0x20 || ch > 0x7e) {
+                if (ch < 0x20 || ch > 0x7E) {
                     String str = "0000" + Integer.toString(ch, 16);
                     retval.append("\\u").append(str.substring(str.length() - 4));
                 } else {
