@@ -1,6 +1,6 @@
 import java.io.ByteArrayOutputStream
 
-val version = "2.12.0"
+val version = "2.12.1"
 val suffix = ""
 
 // Strings embedded into the build.
@@ -103,7 +103,7 @@ subprojects {
                     url = if (suffix.contains("SNAPSHOT")) {
                         uri("https://s01.oss.sonatype.org/content/repositories/snapshots/")
                     } else {
-                        uri("https://s01.oss.sonatype.org/service/local/staging/deploy/maven2/")
+                        uri("https://ossrh-staging-api.central.sonatype.com/service/local/staging/deploy/maven2/")
                     }
                     credentials {
                         username = (project.properties["ossrhUsername"] ?: "").toString()
